/****************************************************************************
 Title	:   HD44780U LCD library
 Author:    Peter Fleury <pfleury@gmx.ch>  http://jump.to/fleury
 Date:	    23-Sep-2000
 Software:  AVR-GCC with AVR-AS
 Target:    any AVR device, memory mapped mode only for AT90S4414/8515/Mega

 DESCRIPTION
       Basic routines for interfacing a HD44780U-based text lcd display

       Based on Volker Oth's lcd library (http://members.xoom.com/volkeroth),
       changed lcd_init(), added additional constants for lcd_command(),
       added 4-bit I/O mode, improved and optimized code.
       Library can be operated in memory mapped mode (LCD_IO_MODE=0) or in 
       4-bit IO port mode (LCD_IO_MODE=1). 8-bit IO port mode not supported.
       
       Memory mapped mode compatible with Kanda STK200, but supports also
       generation of R/W signal through A8 address line.

 USAGE
       See the C include lcd.h file for a description of each function
       
*****************************************************************************/
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "lcd.h"


/* 
** constants/macros 
*/

#define PIN(x) (*(&x - 2))  /* address of data direction register of port x */
#define DDR(x) (*(&x - 1))  /* address of input register of port x          */


#if LCD_IO_MODE
//#define lcd_e_delay() __asm__ __volatile__( "rjmp 1f\n 1:" );  /* delay 500ns with 4Mhz */
#define lcd_e_delay() __asm__ __volatile__( "nop; nop; nop; nop; nop; nop;" );  /* delay 500ns with 4Mhz */

#define lcd_e_high()    sbi(LCD_E_PORT, LCD_E_PIN)
#define lcd_e_low()     cbi(LCD_E_PORT, LCD_E_PIN)
#define lcd_e_toggle()  toggle_e()
#endif

#if LCD_IO_MODE
#if LCD_LINES==1
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_1LINE 
#else
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_2LINES 
#endif
#else
#if LCD_LINES==1
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_8BIT_1LINE
#else
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_8BIT_2LINES
#endif
#endif


/* 
** function prototypes 
*/
static void delay(u16 us);
#if LCD_IO_MODE
static void toggle_e(void);
#endif

/*
** local functions
*/

static void delay(u16 us) 
/* delay for a minimum of <us> microseconds    */
/* with a 4Mhz crystal, the resolution is 1 us */
{
    while ( us ) us--;
} 


#if LCD_IO_MODE
static void toggle_e(void)
/* toggle Enable Pin */
{
    lcd_e_high();
    lcd_e_delay();
    lcd_e_low();
}
#endif

#define STOPINT 


static void inline lcd_write(u08 data,u08 rs) 
{
    /* configure data pins as output */
  //  outp(0xFF, DDR(LCD_DATA_PORT));

  // int ausschalten!
  DDR(LCD_DATA_PORT) = 0x0F; // da ist auch noch der OE fr die Relais dabei !

  cbi(LCD_RW_PORT, LCD_RW_PIN);

  if (rs) {   /* write data        (RS=1, RW=0) */
    sbi(LCD_RS_PORT, LCD_RS_PIN);
  } else {    /* write instruction (RS=0, RW=0) */
    cbi(LCD_RS_PORT, LCD_RS_PIN);
    }
  /* output high nibble first */

  LCD_DATA_PORT = ((data >> 4) & 0x0F) | 0xF0;
  lcd_e_toggle();
  
  /* output low nibble */
  LCD_DATA_PORT = (data  & 0x0F) | 0xF0;
  lcd_e_toggle();

  // inactive (Pullups)
  LCD_DATA_PORT = 0xFF;

  // hier den Int wieder ein schalten
}



static u08 lcd_read(u08 rs) 
{
    register u08 dataH, dataL;


    if (rs) sbi(LCD_RS_PORT, LCD_RS_PIN);    /* RS=1: read data      */
    else    cbi(LCD_RS_PORT, LCD_RS_PIN);    /* RS=0: read busy flag */
    sbi(LCD_RW_PORT, LCD_RW_PIN);            /* RW=1  read mode      */

    /* configure data pins as input */
    outp(0x00, DDR(LCD_DATA_PORT));
    
    lcd_e_high();
    lcd_e_delay();
    dataH = inp(PIN(LCD_DATA_PORT));         /* read high nibble first */
    lcd_e_low();
    
    lcd_e_delay();                           /* Enable 500ns low       */
    
    lcd_e_high();
    lcd_e_delay();
    dataL = inp(PIN(LCD_DATA_PORT));         /* read low nibble        */
    lcd_e_low();

    
    return ( ((dataH<<4) &0xF0) | (dataL &0x0F) );
}



static inline unsigned char lcd_waitbusy(void)
/* loops while lcd is busy, reads address counter */
{
    register unsigned char c;
    
    while ( (c=lcd_read(0)) & (1<<LCD_BUSY)) {}
    //delay(1000);
    return (c);  // return address counter
}


static inline void lcd_newline(u08 pos)
/* goto start of next line */
{
    register u08 addressCounter;


#if LCD_LINES==1
    addressCounter = 0;
#endif
#if LCD_LINES==2
    if ( pos < (LCD_START_LINE2) )
        addressCounter = LCD_START_LINE2;
    else
        addressCounter = LCD_START_LINE1;
#endif
#if LCD_LINES==4
    if ( pos < LCD_START_LINE3 )
        addressCounter = LCD_START_LINE2;
    else if ( (pos >= LCD_START_LINE2) && (pos < LCD_START_LINE4) )
        addressCounter = LCD_START_LINE3;
    else if ( (pos >= LCD_START_LINE3) && (pos < LCD_START_LINE2) )
        addressCounter = LCD_START_LINE4;
    else 
        addressCounter = LCD_START_LINE1;
#endif
    lcd_command((1<<LCD_DDRAM)+addressCounter);

}/* lcd_newline */


/*
** PUBLIC FUNCTIONS 
*/

void lcd_command(u08 cmd)
/* send commando <cmd> to LCD */
{
    lcd_waitbusy();
    lcd_write(cmd,0);
}


void lcd_gotoxy(u08 x, u08 y)
/* goto position (x,y) */
{
#if LCD_LINES==1
    lcd_command((1<<LCD_DDRAM)+LCD_START_LINE1+x);
#endif
#if LCD_LINES==2
    if ( y==0 ) 
        lcd_command((1<<LCD_DDRAM)+LCD_START_LINE1+x);
    else
        lcd_command((1<<LCD_DDRAM)+LCD_START_LINE2+x);
#endif
#if LCD_LINES==4
    if ( y==0 )
        lcd_command((1<<LCD_DDRAM)+LCD_START_LINE1+x);
    else if ( y==1)
        lcd_command((1<<LCD_DDRAM)+LCD_START_LINE2+x);
    else if ( y==2)
        lcd_command((1<<LCD_DDRAM)+LCD_START_LINE3+x);
    else /* y==3 */
        lcd_command((1<<LCD_DDRAM)+LCD_START_LINE4+x);
#endif

}/* lcd_gotoxy */


void lcd_clrscr(void)
/* clear lcd and set cursor to home position */
{
    lcd_command(1<<LCD_CLR);
}


void lcd_home(void)
/* set cursor to home position */
{
    lcd_command(1<<LCD_HOME);
}


void lcd_putc(char c)
/* print character at current cursor position */
{
    register unsigned char pos;
    
    pos = lcd_waitbusy();   // read busy-flag and address counter
    if (c=='\n')
        lcd_newline(pos);
    else
        lcd_write(c, 1);
}


void lcd_puts(const char *s)
/* print string on lcd (no auto linefeed) */
{
    register char c;

    while ( (c = *s++) ) {
        lcd_putc(c);
    }

}


void lcd_puts_p(const char *progmem_s)
/* print string from program memory on lcd (no auto linefeed) */
{
    register char c;

    while ( (c = PRG_RDB(progmem_s++)) ) {
        lcd_putc(c);
    }

}


void lcd_init(u08 dispAttr)
/* initialize display and select type of cursor */
/* dispAttr: LCD_DISP_OFF, LCD_DISP_ON, LCD_DISP_ON_CURSOR, LCD_DISP_CURSOR_BLINK */
{

    /*------ Initialize lcd to 4 bit i/o mode -------*/
    
    outp( 0x0F, DDR(LCD_DATA_PORT) );         /* all port bits as output */
    sbi(DDR(LCD_RS_PORT), LCD_RS_PIN);
    sbi(DDR(LCD_RW_PORT), LCD_RW_PIN);
    sbi(DDR(LCD_E_PORT), LCD_E_PIN);

    sbi(LCD_RS_PORT, LCD_RS_PIN);
    sbi(LCD_RW_PORT, LCD_RW_PIN);
    sbi(LCD_E_PORT, LCD_E_PIN);
    delay(16000);        /* wait 16ms or more after power-on       */
    
    /* initial write to lcd is 8bit */
    outp(((LCD_FUNCTION_8BIT_1LINE >> 4) | 0xF0),LCD_DATA_PORT);
    lcd_e_toggle();
    delay(4992);         /* delay, busy flag can't be checked here */
    
    outp(((LCD_FUNCTION_8BIT_1LINE >> 4) | 0xF0),LCD_DATA_PORT);
    lcd_e_toggle();
    delay(64);           /* delay, busy flag can't be checked here */
    
    outp(((LCD_FUNCTION_8BIT_1LINE >> 4) | 0xF0),LCD_DATA_PORT);
    lcd_e_toggle();
    delay(64);           /* delay, busy flag can't be checked here */
    
    outp(((LCD_FUNCTION_4BIT_1LINE >> 4) | 0xF0),LCD_DATA_PORT); /* set IO mode to 4bit */
    lcd_e_toggle();
    
    /* from now the lcd only accepts 4 bit I/O, we can use lcd_command() */    

    lcd_command(LCD_FUNCTION_DEFAULT);      /* function set: display lines  */
    lcd_command(LCD_DISP_OFF);              /* display off                  */
    lcd_clrscr();                           /* display clear                */ 
    lcd_command(LCD_MODE_DEFAULT);          /* set entry mode               */
    lcd_command(dispAttr);                  /* display/cursor control       */

    outp( 0x0F, DDR(LCD_DATA_PORT) );         /* input ready for keys */
}/* lcd_init */
